@@ -1,11 +1,13 @@
 from visual_kinematics.Frame import *
 import matplotlib.pyplot as plt
-from scipy.optimize import least_squares
+from math import pi
+# from scipy.optimize import least_squares
 from mpl_toolkits.mplot3d import Axes3D
 from matplotlib.widgets import Slider
 
 import logging
 
+
 class Robot:
     # ==================
     # dh_params [n, 4]
@@ -15,12 +17,13 @@ class Robot:
     # | ... |   ...   | ... |   ...   |
     # ==================
     def __init__(self, dh_params, dh_type="normal", analytical_inv=None
-                 , plot_xlim=[-0.5, 0.5], plot_ylim=[-0.5, 0.5], plot_zlim=[0.0, 1.0]):
+                 , plot_xlim=[-0.5, 0.5], plot_ylim=[-0.5, 0.5], plot_zlim=[0.0, 1.0]
+                 , inv_m = "jac_pinv", step_size=5e-1, max_iter=300, final_loss=1e-4):
         self.dh_params = dh_params
         self.dh_type = dh_type # can be only normal or modified
         if dh_type != "normal" and dh_type != "modified":
             raise Exception("dh_type can only be \"normal\" or \"modified\"")
-        self.initial_offset = self.dh_params[:, 1:2].copy()
+        self.initial_offset = self.dh_params[:, 1].copy()
         self.analytical_inv = analytical_inv
         # plot related
         self.plot_xlim = plot_xlim
@@ -28,6 +31,13 @@ class Robot:
         self.plot_zlim = plot_zlim
         self.figure = plt.figure()
         self.ax = self.figure.add_subplot(111, projection="3d")
+        # inverse settings
+        self.inv_m = inv_m
+        if inv_m != "jac_t" and inv_m != "jac_pinv":
+            raise Exception("Motion type can only be \"jac_t\" or \"jac_inv\"!")
+        self.step_size = step_size
+        self.max_iter = max_iter
+        self.final_loss = final_loss
 
     @property
     def num_axis(self):
@@ -35,7 +45,7 @@ class Robot:
 
     @property
     def axis_values(self):
-        return self.dh_params[:, 1:2].reshape([self.num_axis, ]) - self.initial_offset.reshape([self.num_axis, ])
+        return self.dh_params[:, 1] - self.initial_offset
 
     # transformation between axes
     @property
@@ -63,8 +73,24 @@ class Robot:
     def end_frame(self):
         return self.axis_frames[-1]
 
+    @property
+    def jacobian(self):
+        axis_fs = self.axis_frames
+        jac = np.zeros([6, self.num_axis])
+        if self.dh_type == "normal":
+            jac[0:3, 0] = np.cross(np.array([0., 0., 1.]), axis_fs[-1].t_3_1.reshape([3, ]))
+            jac[3:6, 0] = np.array([0., 0., 1.])
+            for i in range(1, self.num_axis):
+                jac[0:3, i] = np.cross(axis_fs[i-1].z_3_1.reshape([3, ]), (axis_fs[-1].t_3_1 - axis_fs[i-1].t_3_1).reshape([3, ]))
+                jac[3:6, i] = axis_fs[i-1].z_3_1.reshape([3, ])
+        if self.dh_type == "modified":
+            for i in range(0, self.num_axis):
+                jac[0:3, i] = np.cross(axis_fs[i].z_3_1.reshape([3, ]), (axis_fs[-1].t_3_1 - axis_fs[i].t_3_1).reshape([3, ]))
+                jac[3:6, i] = axis_fs[i].z_3_1.reshape([3, ])
+        return jac
+
     def forward(self, theta_x):
-        self.dh_params[:, 1:2] = theta_x.reshape([self.num_axis, 1]) + self.initial_offset
+        self.dh_params[:, 1] = theta_x + self.initial_offset
         return self.end_frame
 
     def inverse(self, end_frame):
@@ -75,21 +101,32 @@ class Robot:
 
     def inverse_analytical(self, end_frame, method):
         theta_x = method(self.dh_params, end_frame)
+        self.forward(theta_x)
         return theta_x
 
     def inverse_numerical(self, end_frame):
-        theta_x = self.axis_values
-        ls = least_squares(Robot.cost_inverse, theta_x, args=(self, end_frame), ftol=1e-15)
-        if ls.cost > 1e-4:
-            logging.error("The pose is out of robot's reach!!!")
-
-    @staticmethod
-    def cost_inverse(x, robot, end_frame):
-        end = robot.forward(x)
-        # residual = np.zeros([6, ])
-        # residual[0:3] = (end.t_3_1 - end_frame.t_3_1).reshape([3, ])
-        # residual[3:6] = (end.r_3 - end_frame.r_3).reshape([3, ])
-        return (end.t_4_4 - end_frame.t_4_4).reshape([16, ])
+        last_dx = np.zeros([6, 1])
+        for _ in range(self.max_iter):
+            if self.inv_m == "jac_t":
+                jac = self.jacobian.T
+            else:
+                jac = np.linalg.pinv(self.jacobian)
+            end = self.end_frame
+            dx = np.zeros([6, 1])
+            dx[0:3, 0] = (end_frame.t_3_1 - end.t_3_1).reshape([3, ])
+            diff = end.inv * end_frame
+            dx[3:6, 0] = end.r_3_3.dot(diff.r_3.reshape([3, 1])).reshape([3, ])
+            if np.linalg.norm(dx, ord=2) < self.final_loss or np.linalg.norm(dx - last_dx, ord=2) < 0.1*self.final_loss:
+                for i in range(self.num_axis):
+                    while self.dh_params[i, 1] > pi:
+                        self.dh_params[i, 1] -= 2*pi
+                    while self.dh_params[i, 1] < -pi:
+                        self.dh_params[i, 1] += 2*pi
+                return self.axis_values
+            dq = self.step_size * jac.dot(dx)
+            self.forward(self.axis_values + dq.reshape([self.num_axis, ]))
+            last_dx = dx
+        logging.error("Pose cannot be reached!")
 
     # ===============
     # trajectory
@@ -112,7 +149,7 @@ class Robot:
         tra_array = np.zeros([len(tra), self.num_axis])
         for i in range(len(tra)):
             if motion == "p2p":
-                self.inverse_numerical(tra[i])
+                self.inverse(tra[i])
                 tra_array[i] = self.axis_values
             if motion == "lin":
                 tra_array[i, 0:3] = np.array(tra[i].t_3_1.reshape([3, ]))
@@ -134,7 +171,7 @@ class Robot:
                 inter_values[progress] = tra_array[index] * (1-p_temp) + tra_array[index+1] * p_temp
             if motion == "lin":
                 xyzabc = tra_array[index] * (1-p_temp) + tra_array[index+1] * p_temp
-                self.inverse_numerical(Frame.from_euler_3(xyzabc[3:6], xyzabc[0:3].reshape([3, 1])))
+                self.inverse(Frame.from_euler_3(xyzabc[3:6], xyzabc[0:3].reshape([3, 1])))
                 inter_values[progress] = self.axis_values
         return inter_values
 
@@ -203,75 +240,4 @@ class Robot:
         self.draw()
         self.ax.plot_wireframe(x, y, np.array([z]), color="lightblue")
         plt.show()
-
-    def show_trajectory_old(self, tra, motion="p2p"):
-        if motion != "p2p" and motion != "lin":
-            logging.warning("Motion type can only be \"p2p\" or \"lin\"!")
-        if len(tra) < 2:
-            logging.warning("Please give at least 2 frames for the trajectory!")
-
-        # setup slider
-        axamp = plt.axes([0.15, .06, 0.75, 0.02])
-        samp = Slider(axamp, "progress", 0., 1., valinit=0)
-
-        # interpolation
-        # !!! currently use linear interpolation
-        # !!! currently based on only the length of each segment, future work will take orientation also into consideration
-        # !!! currently all segment must share same method
-
-        tra_array = np.zeros([len(tra), self.num_axis])
-        # axis angles for p2p, xyzabc for lin
-        for i in range(len(tra)):
-            if motion == "p2p":
-                self.inverse_numerical(tra[i])
-                tra_array[i] = self.axis_values
-            if motion == "lin":
-                tra_array[i, 0:3] = np.array(tra[i].t_3_1.reshape([3, ]))
-                tra_array[i, 3:6] = np.array(tra[i].euler_3)
-
-        length = []
-        length_total = 0.
-        for i in range(len(tra)-1):
-            length.append(tra[i].distance_to(tra[i+1]))
-            length_total += length[i]
-
-        def move_to_interval(progress):
-            index = 0
-            p_temp = progress
-            for i in range(len(length)):
-                if p_temp - length[i] > 1e-5: # prevent numerical error
-                    p_temp -= length[i]
-                    index += 1
-                else:
-                    break
-            p_temp /= length[index]
-            if motion == "p2p":
-                self.forward(tra_array[index] * (1-p_temp) + tra_array[index+1] * p_temp)
-            if motion == "lin":
-                xyzabc = tra_array[index] * (1-p_temp) + tra_array[index+1] * p_temp
-                self.inverse_numerical(Frame.from_euler_3(xyzabc[3:6], xyzabc[0:3].reshape([3, 1])))
-
-        # save point for drawing trajectory
-        x, y, z = [], [], []
-        for i in range(101):
-            move_to_interval(i * length_total / 100.)
-            x.append(self.end_frame.t_3_1[0, 0])
-            y.append(self.end_frame.t_3_1[1, 0])
-            z.append(self.end_frame.t_3_1[2, 0])
-
-        def update(val):
-            move_to_interval(samp.val * length_total)
-            self.draw()
-            # plot trajectory
-            self.ax.plot_wireframe(x, y, np.array([z]), color="lightblue")
-            self.figure.canvas.draw_idle()
-
-        samp.on_changed(update)
-
-        # plot initial
-        self.inverse_numerical(tra[0])
-        self.inverse_numerical(tra[1])
-        self.inverse_numerical(tra[0])
-        self.draw()
-        self.ax.plot_wireframe(x, y, np.array([z]), color="lightblue")
-        plt.show()
+        return inter_values
